import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat, einsum
import math
from typing import Optional, Tuple


class Base(nn.Module):
    def __init__(self, divide=1.0, learned=False):
        super().__init__()
        self.learned = learned
        if not learned:
            self.div = divide
        else:
            self.div = nn.Parameter(torch.tensor(math.log(divide)))

        self.forward = self.mod_x

    def mod_x(self, x, pos=None, **kwargs):
        return x

    def mod_qkv(self, q, k, v, pos=None, **kwargs):
        return q, k, v

    def mod_mask(self, mask, q, k, v, pos=None, **kwargs):
        return mask

    def apply_pos_division(self, x):
        if not self.learned:
            return x / self.div
        else:
            return x / self.div.exp().clamp(min=1e-8)

    @staticmethod
    def default_pos_x(x):
        return torch.arange(x.size(-2), device=x.device).expand(*x.shape[:-1])


class Sinusoidal(Base):
    """Sinusoidal positional embedding as in Vaswani et al. (2017)
    Supports specifying positions, masking, and division of positional range.

    Parameters
    ----------
    dim : int
        Hidden size of the embeddings
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(self, dim: int, divide: float = 1.0, learned_div: bool = False):
        super().__init__(divide=divide, learned=learned_div)
        self.dim = dim
        inv_freq = 1 / (10000 ** (torch.arange(0.0, dim, 2) / dim))
        self.register_buffer("inv_freq", inv_freq)

    def mod_x(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify X

        Parameters
        ----------
        x : torch.Tensor
            (B,*,L,H)
        pos : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None for computing positions from 0 to L-1
            If sequence length is smaller than x, will pad tokens on the lefthand side to not have any positional encodings added.
        mask : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None
            A boolean mask can be used to explicitly indicate which positions should not have positional encodings added.
            If sequence length is smaller than x, will pad tokens on the lefthand side to have a positional encoding added.

        Returns
        -------
        torch.Tensor
            (B,*,L,H)
        """
        shp = list(x.shape)
        l = shp[-2]
        assert shp[-1] == self.dim

        pos_emb = self.default_pos_x(x).to(self.inv_freq) if pos is None else pos
        shp[-2] = pos_emb.shape[-1]

        sinusoid_inp = self.apply_pos_division(pos_emb.unsqueeze(-1) * self.inv_freq)
        pos_emb = (
            torch.stack((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)
            .view(*shp)
            .to(x.dtype)
        )

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2] - mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, l - shp[-2], 0))

        return x + pos_emb


class LearnedVocab(Base):
    """Learned vocab as in Devlin et al. (2018)
    Supports specifying positions.
    Only works for discrete positional indices.

    Parameters
    ----------
    dim : int
        Hidden size of the embeddings
    max_seq_len : int
        Maximum sequence length or vocab size of the learned embeddings
    """

    def __init__(self, dim: int, max_seq_len: int):
        super().__init__()
        self.emb = nn.Embedding(max_seq_len, dim)

    def mod_x(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify X

        Parameters
        ----------
        x : torch.Tensor
            (B,*,L,H)
        pos : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None for computing positions from 0 to L-1
            If sequence length is smaller than x, will pad tokens on the lefthand side to not have any positional encodings added.
        mask : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None
            A boolean mask can be used to explicitly indicate which positions should not have positional encodings added.
            If sequence length is smaller than x, will pad tokens on the lefthand side to have a positional encoding added.

        Returns
        -------
        torch.Tensor
            (B,*,L,H)
        """
        shp = list(x.shape)
        assert shp[-1] == self.emb.embedding_dim
        pos = self.default_pos_x(x).long() if pos is None else pos

        pos_emb = self.emb(pos)

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2] - mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        l = pos_emb.shape[-2]
        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, shp[-2] - l, 0))

        return x + pos_emb


class LearnedContinuous(Base):
    """Learned embeddings with a continuity between absolute positional indices, as learned by a series of linear layers.
    Supports specifying positions, and division of positional range.

    Parameters
    ----------
    dim : int
        Hidden size of the embeddings
    depth : int, optional
        Number of hidden layers in the positional embedding network.
        Will follow this structure:
        Linear -> (Norm if norm) { -> Swish -> Linear -> (Norm if norm) } * (depth-1)
        By default 1, for a linear embedding.
    norm : bool, optional
        Whether to use LayerNorms in the embedding net, by default False
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(
        self,
        dim: int,
        depth: int = 1,
        norm: bool = False,
        divide: float = 1.0,
        learned_div: bool = False,
    ):
        super().__init__(divide=divide, learned=learned_div)

        self.mlp = nn.ModuleList([])
        self.mlp.append(
            nn.Sequential(
                nn.Linear(1, dim),
                nn.LayerNorm(dim) if norm else nn.Identity(),
            )
        )

        for _ in range(depth - 1):
            self.mlp.append(
                nn.Sequential(
                    nn.SiLU(),
                    nn.Linear(dim, dim),
                    nn.LayerNorm(dim) if norm else nn.Identity(),
                )
            )

    def mod_x(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify X

        Parameters
        ----------
        x : torch.Tensor
            (B,*,L,H)
        pos : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None for computing positions from 0 to L-1
            If sequence length is smaller than x, will pad tokens on the lefthand side to not have any positional encodings added.
        mask : Optional[torch.Tensor], optional
            (B,*,L) or (B,*,L-x), by default None
            A boolean mask can be used to explicitly indicate which positions should not have positional encodings added.
            If sequence length is smaller than x, will pad tokens on the lefthand side to have a positional encoding added.

        Returns
        -------
        torch.Tensor
            (B,*,L,H)
        """
        shp = list(x.shape)

        pos = self.default_pos_x(x).to(x) if pos is None else pos

        pos_emb = self.apply_pos_division(pos).unsqueeze(-1)
        for layer in self.mlp:
            pos_emb = layer(pos_emb)

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2] - mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        l = pos_emb.shape[-2]
        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, shp[-2] - l, 0))

        return x + pos_emb


class Rotary(Base):
    """Rotary embedding as in RoFormer / RoPE
    Supports specifying positions and division of positional range.

    Parameters
    ----------
    head_dim : int
        Hidden dimensions per head
    n_dims : Optional[int], optional
        number of dimensions (per head) to apply rotations on.
            Can be used to control how strong the positional bias should be.
            By default None to use all dims.
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(
        self,
        head_dim: int,
        n_dims: Optional[int] = None,
        divide: float = 1.0,
        learned_div: bool = False,
    ):
        super().__init__(divide=divide, learned=learned_div)

        thetas = 1.0 / (10000 ** (torch.arange(0, head_dim, 2).float() / head_dim))
        if n_dims is not None:
            assert n_dims % 2 == 0
            thetas[(n_dims // 2) :] = 0
        self.register_buffer("thetas", thetas)

    def mod_qkv(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        pos_q_k: Optional[torch.Tensor] = None,
        pos_q: Optional[torch.Tensor] = None,
        pos_k: Optional[torch.Tensor] = None,
        self_attn_mode: bool = True,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Modify Q, K and V

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        pos_q_k : Optional[torch.Tensor], optional
            (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2
            By default None for computing positions from 0 to L-1
        pos_q : Optional[torch.Tensor], optional
            (B, *, L1). Positions of q in cross attention mode
            By default None for computing positions from 0 to L1-1
        pos_k : Optional[torch.Tensor], optional
            (B, *, L2). Positions of k in cross attention mode
            By default None for computing positions from 0 to L2-1
        self_attn_mode : bool, optional
            Whether to use the same positions for q and k (pos_q_k) or use different positions for each (pos_q) and (pos_k)
            by default True

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            q, k, v [(B, *, L1, NH, H), (B, *, L2, NH, H), (B, *, L2, NH, H)]
        """
        if self_attn_mode:
            assert q.shape[-3] == k.shape[-3]
            pos_q_k = self.default_pos_x(q[..., 0, :]) if pos_q_k is None else pos_q_k
            sin, cos = self.get_rotations(pos_q_k)

            l_sin, l_q = sin.shape[-3], q.shape[-3]
            if l_sin != l_q:
                sin, cos = map(
                    lambda t: F.pad(t, (0, 0, 0, 0, l_q - l_sin, 0), (sin, cos))
                )

            q = q * cos.to(q) + self.rotate_every_two(q) * sin.to(q)
            k = k * cos.to(k) + self.rotate_every_two(k) * sin.to(k)
        else:
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k
            sin_q, cos_q = self.get_rotations(pos_q)
            sin_k, cos_k = self.get_rotations(pos_k)

            l_sin, l_q = sin_q.shape[-3], q.shape[-3]
            if l_sin != l_q:
                sin_q, cos_q = map(
                    lambda t: F.pad(t, (0, 0, 0, 0, l_q - l_sin, 0), (sin_q, cos_q))
                )
            l_sin, l_k = sin_k.shape[-3], k.shape[-3]
            if l_sin != l_k:
                sin_k, cos_k = map(
                    lambda t: F.pad(t, (0, 0, 0, 0, l_k - l_sin, 0), (sin_k, cos_k))
                )

            q = q * cos_q.to(q) + self.rotate_every_two(q) * sin_q.to(q)
            k = k * cos_k.to(k) + self.rotate_every_two(k) * sin_k.to(k)

        return q, k, v

    def get_rotations(self, pos):
        mthetas = (
            self.apply_pos_division(pos).to(self.thetas)[..., None] * self.thetas
        )  # B, *, h
        sin, cos = map(
            lambda t: repeat(t, "b ... h  -> b ... (h j)", j=2),
            (mthetas.sin(), mthetas.cos()),
        )
        sin, cos = map(lambda t: t.unsqueeze(-2), (sin, cos))
        return sin, cos

    @staticmethod
    def rotate_every_two(x):
        x = x.clone()
        x = rearrange(x, "... (d j) -> ... d j", j=2)
        x1, x2 = x.unbind(dim=-1)
        x = torch.stack((-x2, x1), dim=-1)
        return rearrange(x, "... d j -> ... (d j)")


class ALiBi(Base):
    """Attention with linear biases as in Press et al. 2022
    Supports specifying positions, division of positional range, using only a fraction of the heads, and asymmetric bias for bidirectional cases.

    Parameters
    ----------
    n_heads : int
        Number of heads
    use_n_heads : bool, optional
        Number of heads to use. Can be used to control how strong the positional bias should be, by default None to use all heads
    asymmetric : bool, optional
        Whether to use assymetric positional biases to differentiate between negative or positive relative positions.
        Implemented according to solution #3 proposed here https://github.com/ofirpress/attention_with_linear_biases/issues/5.
        By default False
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(
        self,
        n_heads: int,
        use_n_heads: bool = None,
        asymmetric: bool = False,
        divide: float = 1.0,
        learned_div: bool = False,
    ):
        super().__init__(divide=divide, learned=learned_div)

        self.asymmetric = asymmetric
        use_n_heads = n_heads if use_n_heads is None else use_n_heads
        if not asymmetric:
            slopes = torch.tensor(
                self._get_slopes(use_n_heads) + [0] * (n_heads - use_n_heads)
            )
        else:
            slopes = self._get_slopes(use_n_heads)
            slopes = [slopes[(i // 2) * 2] for i in range((len(slopes)))]
            slopes = torch.tensor(slopes + [0] * (n_heads - use_n_heads))
        slopes = rearrange(slopes, "h -> h 1 1")
        self.register_buffer("slopes", slopes)

        self.div = divide

    def mod_mask(
        self,
        mask: Optional[torch.Tensor],
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        pos_q_k: Optional[torch.Tensor] = None,
        pos_q: Optional[torch.Tensor] = None,
        pos_k: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify mask

        Parameters
        ----------
        mask : Optional[torch.Tensor]
            (B, *, NH, L1, L2), can pass None for no pre-existing mask.
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        pos_q_k : Optional[torch.Tensor], optional
            (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2
            By default None for using pos_q and pos_k instead
        pos_q : Optional[torch.Tensor], optional
            (B, *, L1). Positions of q in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L1-1
        pos_k : Optional[torch.Tensor], optional
            (B, *, L2). Positions of k in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L2-1

        Returns
        -------
        torch.Tensor
            (B, *, NH, L1, L2)
        """
        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k

        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        relative_pos = pos_q[..., :, None] - pos_k[..., None, :]
        bias = self.apply_pos_division(relative_pos.unsqueeze(-3) * self.slopes)

        if self.asymmetric:
            bias = self.asymmetric_bias(bias)
        bias = torch.abs(bias)

        if mask is not None:
            l1, l2 = q.shape[-3], k.shape[-3]
            mask_l1, mask_l2 = mask.shape[-2], mask.shape[-1]
            if (l1 != mask_l1) or (l2 != mask_l2):
                mask = F.pad(mask, (l2 - mask_l2, 0, l1 - mask_l1, 0))

        return (0 if mask is None else mask) - bias

    @staticmethod
    def _get_slopes(heads):
        def get_slopes_power_of_2(n):
            start = 2 ** (-(2 ** -(math.log2(n) - 3)))
            ratio = start
            return [start * ratio**i for i in range(n)]

        if math.log2(heads).is_integer():
            return get_slopes_power_of_2(heads)

        closest_power_of_2 = 2 ** math.floor(math.log2(heads))
        return (
            get_slopes_power_of_2(closest_power_of_2)
            + get_slopes_power_of_2(2 * closest_power_of_2)[0::2][
                : heads - closest_power_of_2
            ]
        )

    @staticmethod
    def asymmetric_bias(bias):
        bias[:, ::2][bias[:, ::2] > 0] = torch.finfo(bias.dtype).max
        bias[:, 1::2][bias[:, 1::2] < 0] = torch.finfo(bias.dtype).max
        return bias


class DPB(Base):
    """Dynamic positional bias.
    Computes a bias to every element of the attention matrix based on their relative position via a parameterized MLP
    Described in https://arxiv.org/abs/2108.00154, https://arxiv.org/abs/2111.09883,
    https://github.com/lucidrains/x-transformers#dynamic-positional-bias
    Supports specifying positions and division of positional range

    Parameters
    ----------
    dim : int
        number of hidden dimensions of the MLP
    n_heads : int
        Number of heads in the model
    depth : int, optional
        Number of hidden layers in the positional embedding network.
        Will follow this structure:
        Linear -> (Norm if norm) { -> Swish -> Linear -> (Norm if norm) } * (depth-1)
        By default 1, for a linear embedding.
    norm : bool, optional
        Whether to use LayerNorms in the embedding net, by default False
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(
        self,
        dim: int,
        n_heads: int,
        depth: int = 1,
        norm: bool = False,
        divide: float = 1.0,
        learned_div: bool = False,
    ):
        super().__init__(divide=divide, learned=learned_div)
        self.mlp = nn.ModuleList([])
        self.mlp.append(
            nn.Sequential(
                nn.Linear(1, dim),
                nn.LayerNorm(dim) if norm else nn.Identity(),
            )
        )

        for _ in range(depth - 1):
            self.mlp.append(
                nn.Sequential(
                    nn.SiLU(),
                    nn.Linear(dim, dim),
                    nn.LayerNorm(dim) if norm else nn.Identity(),
                )
            )

        self.mlp.append(nn.Linear(dim, n_heads))

        self.div = divide

    def mod_mask(
        self,
        mask: Optional[torch.Tensor],
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        pos_q_k: Optional[torch.Tensor] = None,
        pos_q: Optional[torch.Tensor] = None,
        pos_k: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify mask

        Parameters
        ----------
        mask : Optional[torch.Tensor]
            (B, *, NH, L1, L2), can pass None for no pre-existing mask.
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        pos_q_k : Optional[torch.Tensor], optional
            (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2
            By default None for using pos_q and pos_k instead
        pos_q : Optional[torch.Tensor], optional
            (B, *, L1). Positions of q in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L1-1
        pos_k : Optional[torch.Tensor], optional
            (B, *, L2). Positions of k in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L2-1

        Returns
        -------
        torch.Tensor
            (B, *, NH, L1, L2)
        """
        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k
        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        bias = self.apply_pos_division(pos_q[..., :, None] - pos_k[..., None, :])[
            ..., None
        ].to(q)

        for layer in self.mlp:
            bias = layer(bias)
        bias = bias.transpose(-1, -2).transpose(-2, -3).to(q)

        if mask is not None:
            l1, l2 = q.shape[-3], k.shape[-3]
            mask_l1, mask_l2 = mask.shape[-2], mask.shape[-1]
            if (l1 != mask_l1) or (l2 != mask_l2):
                mask = F.pad(mask, (l2 - mask_l2, 0, l1 - mask_l1, 0))

        return (0 if mask is None else mask) - bias


class XL(Base):
    """Relative positional biases as in Transformer-XL (Dai et al 2019)
    Supports specifying positions and division of positional range

    Parameters
    ----------
    dim : int
        number of hidden of x (total, not per head)
    n_heads : int
        Number of heads in the model
    divide : float, optional
        divide positions by this factor, useful for large (or small) numerical ranges, by default 1.0
    learned_div : bool, optional
        Whether to learn the division factor, if True, div value initialization is as provided by divide argument, by default False
    """

    def __init__(
        self, dim: int, n_heads: int, divide: float = 1.0, learned_div: bool = False
    ):
        super().__init__(divide=divide, learned=learned_div)
        self.embed_lin = nn.Linear(dim, dim)
        self.bias_r_w = self._init_bias(
            nn.Parameter(torch.Tensor(n_heads, dim // n_heads))
        )
        self.bias_r_r = self._init_bias(
            nn.Parameter(torch.Tensor(n_heads, dim // n_heads))
        )
        self.n_heads = n_heads
        self.dim = dim
        inv_freq = 1 / (10000 ** (torch.arange(0.0, dim, 2) / dim))
        self.register_buffer("inv_freq", inv_freq)
        self.div = divide

    def mod_qkv(
        self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Modify Q, K and V

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            q, k, v [(B, *, L1, NH, H), (B, *, L2, NH, H), (B, *, L2, NH, H)]
        """
        return q + self.bias_r_w, k, v

    def mod_mask(
        self,
        mask: Optional[torch.Tensor],
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        pos_q_k: Optional[torch.Tensor] = None,
        pos_q: Optional[torch.Tensor] = None,
        pos_k: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Modify mask

        Parameters
        ----------
        mask : Optional[torch.Tensor]
            (B, *, NH, L1, L2), can pass None for no pre-existing mask.
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        pos_q_k : Optional[torch.Tensor], optional
            (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2
            By default None for using pos_q and pos_k instead
        pos_q : Optional[torch.Tensor], optional
            (B, *, L1). Positions of q in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L1-1
        pos_k : Optional[torch.Tensor], optional
            (B, *, L2). Positions of k in cross attention mode
            Ignored if pos_q_k is specified
            By default None for computing positions from 0 to L2-1

        Returns
        -------
        torch.Tensor
            (B, *, NH, L1, L2)
        """
        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k
        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        relative_pos = (pos_q[..., :, None] - pos_k[..., None, :]).to(
            self.inv_freq.dtype
        )

        sinusoid_inp = self.apply_pos_division(relative_pos[..., None] * self.inv_freq)
        pos_emb = torch.stack((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)

        pos_emb = rearrange(pos_emb, "... d j -> ... (d j)")

        r = rearrange(
            self.embed_lin(pos_emb),
            "... (nh h) -> ... nh h",
            nh=self.n_heads,
            h=q.shape[-1],
        )

        l1, l2 = q.shape[-3], k.shape[-3]
        r_l1, r_l2 = r.shape[-4], r.shape[-3]
        if (l1 != r_l1) or (l2 != r_l2):
            r = F.pad(r, (0, 0, 0, 0, l2 - r_l2, 0, l1 - r_l1, 0))

        q_r = q + self.bias_r_w
        BD = einsum(q_r, r, "... q n h, ... q k n h -> ... n q k")

        if mask is not None:
            l1, l2 = q.shape[-3], k.shape[-3]
            mask_l1, mask_l2 = mask.shape[-2], mask.shape[-1]
            if (l1 != mask_l1) or (l2 != mask_l2):
                mask = F.pad(mask, (l2 - mask_l2, 0, l1 - mask_l1, 0))

        return (0 if mask is None else mask) + BD

    def _init_bias(self, bias):
        bound = 1 / bias.size(1) ** 0.5
        return nn.init.uniform_(bias, -bound, bound)



import torch
from torch import nn, einsum
from einops import rearrange, repeat
import torch.nn.functional as F
import math
from bio_attention import positional
from typing import Optional, Literal, Union


def compl_mod(m, n):
    return int(n * math.ceil(m / n) - m)


class Attention(nn.Module):
    """Scaled-dot product attention operator.
    For more information on kernels: see https://pytorch.org/docs/stable/generated/torch.nn.functional.scaled_dot_product_attention.html

    Parameters
    ----------
    dropout : float, optional
        dropout rate in self attention matrix, by default 0.0
    enable_math : bool, optional
        allow PyTorch C++ implementation, by default True
    enable_flash : bool, optional
        allow FlashAttention implementation, by default True
    enable_mem_efficient : bool, optional
        allow Memory-Efficient implementation, by default True
    """

    def __init__(
        self,
        dropout: float = 0.0,
        enable_math: bool = True,
        enable_flash: bool = True,
        enable_mem_efficient: bool = True,
        **kwargs,
    ):
        super().__init__()
        self.dropout = dropout
        self.context_manager = {
            "enable_math": enable_math,
            "enable_flash": enable_flash,
            "enable_mem_efficient": enable_mem_efficient,
        }
        self.use_context_manager = not all(
            [enable_math, enable_flash, enable_mem_efficient]
        )

    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
    ) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        mask : Optional[torch.Tensor], optional
            (B, *, NH, L1, L2), by default None
        causal : bool, optional
            Perform causal attention. Unlike default pytorch implementation, both a mask and causal can be used jointly, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L1, NH, H)
        """
        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        )  # (B...), NH, L, H
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k")  # (B...), NH, L1, L2

        if causal:
            causal_mask = torch.ones(q.shape[-2], k.shape[-2], dtype=torch.bool)
            causal_mask = causal_mask.triu(diagonal=0).expand(
                q.shape[0], q.shape[1], -1, -1
            )
            mask = (mask if mask is not None else (causal_mask).to(q)).masked_fill(
                causal_mask, -float("inf")
            )

        if self.use_context_manager:
            with torch.backends.cuda.sdp_kernel(**self.context_manager):
                return (
                    F.scaled_dot_product_attention(
                        q,
                        k,
                        v,
                        dropout_p=(self.dropout if self.training else 0),
                        attn_mask=mask,
                        is_causal=False,
                    )
                    .permute(0, 2, 1, 3)
                    .view(*q_shape)
                )
        else:
            return (
                F.scaled_dot_product_attention(
                    q,
                    k,
                    v,
                    dropout_p=(self.dropout if self.training else 0),
                    attn_mask=mask,
                    is_causal=False,
                )
                .permute(0, 2, 1, 3)
                .view(*q_shape)
            )


class RandomAttention(nn.Module):
    """Scaled-dot product attention operator that only randomly attends on a number of keys per query.
    Supports two versions: one that materializes the full matrix and, hence, scales quadratically with sequence length.
    In essence, this is default attention with random masks.
    The other version is memory efficient, scaling linearly with sequence length, but has a lower base efficiency because of the extra steps taken.

    Parameters
    ----------
    n_random_keys : int, optional
        number of keys every query should attend to, by default 64
    dropout : float, optional
        dropout rate in self attention matrix, by default 0.0
    materialize_full : bool, optional
        whether to materialize full attention matrix, by default False
    """

    def __init__(
        self,
        n_random_keys: int = 64,
        dropout: float = 0.0,
        materialize_full: bool = False,
        **kwargs,
    ):
        super().__init__()
        if not materialize_full:
            self.softmax = nn.Softmax(dim=-2)
            self.dropout = nn.Dropout(dropout)

            self.forward = self.forward_indexed
        else:
            self.dropout = dropout
            self.forward = self.forward_naive
        self.n = n_random_keys

    def forward_indexed(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
    ) -> torch.Tensor:
        """Memory-efficient forward pass
        NOTE: for the moment, causal attention with input masks are not implemented.

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        mask : Optional[torch.Tensor], optional
            (B, *, NH, L1, L2), by default None
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L1, NH, H)
        """
        assert not (causal and (mask is not None))

        q_shape = q.shape
        q, k, v = map(lambda t: rearrange(t, "b ... x n h -> (b ...) x n h"), (q, k, v))
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k")

        b, s2, nh, h = k.shape
        s1 = q.shape[1]

        mask = (
            torch.ones(b, nh, s1, s2, dtype=torch.bool).tril(diagonal=0)
            if causal
            else mask
        )
        if mask is not None:
            mask = (
                (~mask).float().masked_fill(~mask, -float("inf"))
                if mask.dtype == torch.bool
                else mask
            )
            mask = mask.to(q)

        q = q * (h**-0.5)

        indices_select = torch.randint(0, s1, (b, s2, self.n)).to(q.device)
        indexer = torch.arange(b).view(b, 1, 1)
        if mask is not None:
            mask = mask.permute(0, 2, 3, 1)[
                indexer, torch.arange(s1)[None, :, None], indices_select
            ].permute(0, 3, 1, 2)

        k = k[indexer, indices_select]
        v = v[indexer, indices_select]

        A = einsum("b q n h, b q k n h -> b q k n", q, k) + (
            0 if mask is None else mask.permute(0, 2, 3, 1)
        )
        A = self.softmax(A)

        A = self.dropout(A)
        z = einsum("b q k n, b q k n h -> b q n h", A, v)

        return z.view(*q_shape)

    def forward_naive(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
    ) -> torch.Tensor:
        """Naive (random masking) forward pass
        NOTE: for the moment, is incompatible with input masks

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        mask : Optional[torch.Tensor], optional
            (B, *, NH, L1, L2), by default None
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L1, NH, H)
        """
        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        )
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k")

        b, nh, s2, h = k.shape
        s1 = q.shape[-2]

        mask = (
            (torch.rand(b, 1, s1, s2) < (self.n / s2))
            .expand(-1, nh, -1, -1)
            .to(k.device)
        )

        return (
            F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=(self.dropout if self.training else 0),
                attn_mask=mask,
                is_causal=causal,
            )
            .permute(0, 2, 1, 3)
            .view(*q_shape)
        )


class WindowAttention(nn.Module):
    """Scaled-dot product attention operator that only attends on a local window of keys per query.
    Supports two versions: one that materializes the full matrix and, hence, scales quadratically with sequence length.
    In essence, this is default attention with a mask.
    The other version is memory efficient, scaling linearly with sequence length, but has a lower base efficiency because of the extra steps taken.

    Parameters
    ----------
    window : int, optional
        Window size, analogous to kernel size in convolutions, should be odd, by default 15
    dropout : float, optional
        dropout rate in self attention matrix, by default 0.0
    materialize_full : bool, optional
        whether to materialize full attention matrix, by default False
    """

    def __init__(
        self,
        window: int = 15,
        dropout: float = 0.0,
        materialize_full: bool = False,
        **kwargs,
    ):
        super().__init__()
        assert window % 2 == 1, "Window size should be an odd integer."

        self.w = int((window - 1) / 2)

        if not materialize_full:
            self.softmax = nn.Softmax(dim=-1)
            self.dropout = nn.Dropout(dropout)
            self.k_ch = window * 2
            self.q_ch = window + 1

            u = torch.triu(torch.full((self.q_ch, self.k_ch), True))
            self.mask = ~torch.logical_and(u, torch.flip(u, [0, 1]))
            self.mask_k_left = torch.clone(self.mask)
            self.mask_k_left[:, : self.w] = True
            self.forward = self.forward_sliced

        else:
            self.dropout = dropout
            self.forward = self.forward_naive

    def forward_sliced(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
    ) -> torch.Tensor:
        """Memory-efficient forward pass
        NOTE: q and k need to have the same sequence length L1 = L2
        NOTE: is incompatible with user-defined masks for the time being.

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        mask : Optional[torch.Tensor], optional
            (B, *, NH, L1, L2), by default None
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L1, NH, H)
        """

        assert mask is None

        assert k.shape[-3] == q.shape[-3], "q and k should have same input length."

        q_shape = q.shape
        q, k, v = map(lambda t: rearrange(t, "b ... x n h -> (b ...) x n h"), (q, k, v))

        b, s, nh, h = k.shape

        q = q * (h**-0.5)

        q_pad = compl_mod(s, self.q_ch)
        k_pad = compl_mod((s + self.w * 2) - self.k_ch, self.q_ch)

        q = F.pad(q, (0,) * 5 + (q_pad,)).unfold(1, self.q_ch, self.q_ch)
        k = F.pad(k, (0,) * 4 + (self.w, self.w + k_pad)).unfold(
            1, self.k_ch, self.q_ch
        )
        v = F.pad(v, (0,) * 4 + (self.w, self.w + k_pad)).unfold(
            1, self.k_ch, self.q_ch
        )

        A = einsum("b c n h q, b c n h k -> b n c q k ", q, k)

        mask_value = -torch.finfo(A.dtype).max
        mask_k_right = torch.clone(self.mask.to(A.device))
        mask_k_right[:, -(self.w + k_pad) :] = True
        if q.shape[1] > 1:
            mask = torch.stack(
                [self.mask_k_left.to(A.device)]
                + [self.mask.to(A.device)] * (q.shape[1] - 2)
                + [mask_k_right]
            )
        else:
            mask = torch.logical_or(
                self.mask_k_left.to(A.device), mask_k_right
            ).unsqueeze(0)
        if causal:
            mask = ~(~mask).tril(diagonal=self.w)

        A.masked_fill_(mask, mask_value)
        A = self.softmax(A)
        A = self.dropout(A)

        z = einsum("b n c q k, b c n h k -> b n c q h", A, v)
        z = z.view(b, nh, -1, h)[:, :, :s].permute(0, 2, 1, 3).view(*q_shape)

        return z

    def forward_naive(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
    ) -> torch.Tensor:
        """Naive (masking) forward pass
        NOTE: q and k need to have the same sequence length L1 = L2
        NOTE: is incompatible with user-defined masks for the time being.

        Parameters
        ----------
        q : torch.Tensor
            (B, *, L1, NH, H)
        k : torch.Tensor
            (B, *, L2, NH, H)
        v : torch.Tensor
            (B, *, L2, NH, H)
        mask : Optional[torch.Tensor], optional
            (B, *, NH, L1, L2), by default None
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L1, NH, H)
        """
        assert mask is None
        assert k.shape[-3] == q.shape[-3], "q and k should have same input length."

        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        )

        b, nh, s2, h = k.shape
        s1 = q.shape[-2]

        u = torch.triu(torch.full((s1, s2), True), diagonal=-self.w)
        mask = (
            torch.logical_and(u, torch.flip(u, [0, 1]))
            .expand(b, nh, -1, -1)
            .to(k.device)
        )
        if causal:
            mask = torch.tril(mask)
        return (
            F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=(self.dropout if self.training else 0),
                attn_mask=mask,
                is_causal=False,
            )
            .permute(0, 2, 1, 3)
            .view(*q_shape)
        )


class AttnLayer(nn.Module):
    """Self-attention layer performing
    (1) projection to q, k and v.
    (2) attention.
    (3) collapsing heads back to same shape as x.
    (4) Final linear project to z.

    Parameters
    ----------
    dim : int
        input and output hidden dimension of x
    attn : Union[Attention, RandomAttention, WindowAttention]
        attention operator. Default, random and windowed attention are implemented.
    nh : int, optional
        number of heads, dim should be divisible by this number, by default 4
    plugin : Optional[positional.Base], optional
        positional bias plugin, for options see the list of implemented biases, by default None
    """

    def __init__(
        self,
        dim: int,
        attn: Union[Attention, RandomAttention, WindowAttention],
        nh: int = 4,
        plugin: Optional[positional.Base] = None,
    ):
        super().__init__()
        assert dim % nh == 0, "dim should be divisible by number of heads"

        self.lin = nn.Linear(dim, 3 * dim)
        self.attn = attn
        self.nh = nh
        self.to_out = nn.Linear(dim, dim)

        self.plugin = plugin

    def forward(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
        **mod_kwargs,
    ) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        x : torch.Tensor
            (B, *, L, H)
        pos : Optional[torch.Tensor], optional
            (B, *, L) or (B, *, L-x), by default None
            If pos has a smaller sequence length than x, it is assumed x has extra tokens added in the beginning of its sequence such as CLS tokens.
            In this case, there is no position for these tokens. Positional biases will make sure those tokens do not partake in positional encoding.
        mask : Optional[torch.Tensor], optional
            By default None, but can be either:
            (1) (B, * L) or (B, * L-x). In this case, expects a boolean mask.
            This type of mask will be copied to (B, * NH, L, L) in a way such that no tokens can attend to positions indicated by False.
            This type of mask will extrapolate CLS tokens to not attend on positions indicated with False.
            (2) (B, * L, L) or (B, * L-x, L-x). In this case, can either be floating point or boolean mask.
            This type of mask will extrapolate CLS token to attend on all tokens.
            For this type of mask, the same mask is applied over all heads
            (3) (B, * NH, L, L) or (B, * NH, L-x, L-x)
            Ditto as previous case, but for this type of mask, different biases/masks can be applied per head.
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L, H)
        """
        q, k, v = torch.split(self.lin(x), x.size(-1), dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "... (n h) -> ... n h", n=self.nh), (q, k, v)
        )  # B, *, L, NH, H

        if mask is not None:
            if q.ndim - mask.ndim == 2:  # B, *, L -> B, *, L, L
                mask = repeat(mask, "... l -> ... (l2) l", l2=mask.shape[-1])
            if q.ndim - mask.ndim == 1:  # B, *, L, L  -> B, *, NH, L, L
                mask = repeat(mask, "... l1 l2 -> ... nh l1 l2", nh=q.shape[-2])

            assert mask.shape[:-3] == q.shape[:-3]
            if mask.dtype == torch.bool:
                mask = (~mask).to(q).masked_fill(~mask, -float("inf"))

            mask = F.pad(
                mask,
                (k.shape[-3] - mask.shape[-1], 0, (q.shape[-3] - mask.shape[-2]), 0),
                value=0,
            )

        mod_kwargs |= {"pos_q_k": pos}

        q, k, v = self.plugin.mod_qkv(q, k, v, **mod_kwargs)

        mask = self.plugin.mod_mask(mask, q, k, v, **mod_kwargs)

        return self.to_out(
            rearrange(
                self.attn(q, k, v, mask=mask, causal=causal),
                "... n h -> ... (n h)",
                n=self.nh,
            )
        )


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x


class GLU(nn.Module):
    def __init__(self, dim, ff_dim, activation):
        super().__init__()
        self.act = activation
        self.proj = nn.Linear(dim, ff_dim * 2)

    def forward(self, x):
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * self.act(gate)


class TransformerLayer(nn.Module):
    """Transformer Layer.
    Performs pre-norm, attention, residual around both, then layernorm + feedforward with residual connection wrapped around it.

    Parameters
    ----------
    attn : Union[Attention, RandomAttention, WindowAttention]
        attention operator. Default, random and windowed attention are implemented.
    dim : int
        input and output hidden dimension of x
    nh : int
        number of heads, dim should be divisible by this number, by default 4
    plugin : Optional[positional.Base], optional
        positional bias plugin, for options see the list of implemented biases, by default None
    dropout : float, optional
        dropout in feedforward, by default 0.2
    glu_ff : bool, optional
        whether to use gated linear feedforward network, by default True
    activation : Literal["relu", "gelu", "swish"], optional
        activation, by default "swish"
    """

    def __init__(
        self,
        attn: Union[Attention, RandomAttention, WindowAttention],
        dim: int,
        nh: int,
        plugin: Optional[positional.Base] = None,
        dropout: float = 0.2,
        glu_ff: bool = True,
        activation: Literal["relu", "gelu", "swish"] = "swish",
    ):
        super().__init__()
        if activation.lower() == "relu":
            act = nn.ReLU()
        elif activation.lower() == "gelu":
            act = nn.GELU()
        elif activation.lower() == "swish":
            act = nn.SiLU()

        self.norm = nn.LayerNorm(dim)

        self.plugin = plugin if plugin is not None else positional.Base()
        self.attn = AttnLayer(dim, attn, nh=nh, plugin=self.plugin)

        project_in = (
            nn.Sequential(nn.Linear(dim, 4 * dim), act)
            if not glu_ff
            else GLU(dim, 4 * dim, act)
        )
        self.ff = Residual(
            nn.Sequential(
                nn.LayerNorm(dim),
                project_in,
                nn.Dropout(dropout),
                nn.Linear(4 * dim, dim),
            )
        )

    def forward(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
        **mod_kwargs,
    ) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        x : torch.Tensor
            (B, *, L, H)
        pos : Optional[torch.Tensor], optional
            (B, *, L) or (B, *, L-x), by default None
            If pos has a smaller sequence length than x, it is assumed x has extra tokens added in the beginning of its sequence such as CLS tokens.
            In this case, there is no position for these tokens. Positional biases will make sure those tokens do not partake in positional encoding.
        mask : Optional[torch.Tensor], optional
            By default None, but can be either:
            (1) (B, * L) or (B, * L-x). In this case, expects a boolean mask.
            This type of mask will be copied to (B, * NH, L, L) in a way such that no tokens can attend to positions indicated by False.
            This type of mask will extrapolate CLS tokens to not attend on positions indicated with False.
            (2) (B, * L, L) or (B, * L-x, L-x). In this case, can either be floating point or boolean mask.
            This type of mask will extrapolate CLS token to attend on all tokens.
            For this type of mask, the same mask is applied over all heads
            (3) (B, * NH, L, L) or (B, * NH, L-x, L-x)
            Ditto as previous case, but for this type of mask, different biases/masks can be applied per head.
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L, H)
        """
        use_mask_to_mod_x = False
        if (mask is not None) and (mask.ndim == x.ndim - 1):
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (x.shape[-2] - mask.shape[-1], 0), value=True)
            use_mask_to_mod_x = True

        x = self.plugin.mod_x(
            x, pos=pos, mask=(mask if use_mask_to_mod_x else None), **mod_kwargs
        )

        x = self.attn(self.norm(x), pos=pos, mask=mask, causal=causal, **mod_kwargs) + x
        return self.ff(x)


class Transformer(nn.Module):
    """Transformer network chaining multiple transformer layers

    Parameters
    ----------
    depth : int
        number of transformer blocks to use
    dim : int
        input and output hidden dimension of x
    nh : int
        number of heads, dim should be divisible by this number
    attentiontype : Literal["vanilla", "random", "window"], optional
        attention operator, by default "vanilla"
    attention_args : dict, optional
        args passed to the attention operator init, by default {}
    plugintype : Literal["none", "sinusoidal", "learned", "learnedcont", "rotary", "ALiBi", "DPB", "XL"], optional
        positional bias plugin, by default "none"
    plugin_args : dict, optional
        arguments passed to positional bias init, by default {}
    only_apply_plugin_at_first : bool, optional
        only apply positional bias at the first layer, by default False
    dropout : float, optional
        dropout in feedforward layers. Take note that attention matrix dropout is controlled via attention_args, by default 0.2
    glu_ff : bool, optional
        whether to use gated linear feedforward network, by default True
    activation : Literal["relu", "gelu", "swish"], optional
        activation, by default "swish"
    """

    def __init__(
        self,
        depth: int,
        dim: int,
        nh: int,
        attentiontype: Literal["vanilla", "random", "window"] = "vanilla",
        attention_args: dict = {},
        plugintype: Literal[
            "none",
            "sinusoidal",
            "learned",
            "learnedcont",
            "rotary",
            "ALiBi",
            "DPB",
            "XL",
        ] = "none",
        plugin_args: dict = {},
        only_apply_plugin_at_first: bool = False,
        dropout: float = 0.2,
        glu_ff: bool = True,
        activation: Literal["relu", "gelu", "swish"] = "swish",
    ):
        super().__init__()

        assert attentiontype in ["vanilla", "random", "window"]

        if attentiontype == "vanilla":
            attn_op = Attention(**attention_args)
        elif attentiontype == "random":
            attn_op = RandomAttention(**attention_args)
        elif attentiontype == "window":
            attn_op = WindowAttention(**attention_args)

        if plugintype == "none":
            plugin = positional.Base()
        elif plugintype == "sinusoidal":
            plugin = positional.Sinusoidal(**plugin_args)
        elif plugintype == "learned":
            plugin = positional.LearnedVocab(**plugin_args)
        elif plugintype == "learnedcont":
            plugin = positional.LearnedContinuous(**plugin_args)
        elif plugintype == "rotary":
            plugin = positional.Rotary(**plugin_args)
        elif plugintype == "ALiBi":
            plugin = positional.ALiBi(**plugin_args)
        elif plugintype == "DPB":
            plugin = positional.DPB(**plugin_args)
        elif plugintype == "XL":
            plugin = positional.XL(**plugin_args)

        layers = []
        layers.append(
            TransformerLayer(
                attn_op,
                dim,
                nh,
                plugin=plugin,
                dropout=dropout,
                glu_ff=glu_ff,
                activation=activation,
            )
        )

        for _ in range(depth - 1):
            layers.append(
                TransformerLayer(
                    attn_op,
                    dim,
                    nh,
                    plugin=(None if only_apply_plugin_at_first else plugin),
                    dropout=dropout,
                    glu_ff=glu_ff,
                    activation=activation,
                )
            )

        self.layers = nn.ModuleList(layers)

    def forward(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        causal: bool = False,
        **mod_kwargs,
    ) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        x : torch.Tensor
            (B, *, L, H)
        pos : Optional[torch.Tensor], optional
            (B, *, L) or (B, *, L-x), by default None
            If pos has a smaller sequence length than x, it is assumed x has extra tokens added in the beginning of its sequence such as CLS tokens.
            In this case, there is no position for these tokens. Positional biases will make sure those tokens do not partake in positional encoding.
        mask : Optional[torch.Tensor], optional
            By default None, but can be either:
            (1) (B, * L) or (B, * L-x). In this case, expects a boolean mask.
            This type of mask will be copied to (B, * NH, L, L) in a way such that no tokens can attend to positions indicated by False.
            This type of mask will extrapolate CLS tokens to not attend on positions indicated with False.
            (2) (B, * L, L) or (B, * L-x, L-x). In this case, can either be floating point or boolean mask.
            This type of mask will extrapolate CLS token to attend on all tokens.
            For this type of mask, the same mask is applied over all heads
            (3) (B, * NH, L, L) or (B, * NH, L-x, L-x)
            Ditto as previous case, but for this type of mask, different biases/masks can be applied per head.
        causal : bool, optional
            Perform causal attention, by default False

        Returns
        -------
        torch.Tensor
            (B, *, L, H)
        """
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=causal, **mod_kwargs)
        return x


class TransformerEncoder(Transformer):
    """Transformer Encoder network chaining multiple transformer layers
    Equal to Transformer, only difference is that Causal = False is automatically decided in forward pass

    Parameters
    ----------
    depth : int
        number of transformer blocks to use
    dim : int
        input and output hidden dimension of x
    nh : int
        number of heads, dim should be divisible by this number
    attentiontype : Literal["vanilla", "random", "window"], optional
        attention operator, by default "vanilla"
    attention_args : dict, optional
        args passed to the attention operator init, by default {}
    plugintype : Literal["none", "sinusoidal", "learned", "learnedcont", "rotary", "ALiBi", "DPB", "XL"], optional
        positional bias plugin, by default "none"
    plugin_args : dict, optional
        arguments passed to positional bias init, by default {}
    only_apply_plugin_at_first : bool, optional
        only apply positional bias at the first layer, by default False
    dropout : float, optional
        dropout in feedforward layers. Take note that attention matrix dropout is controlled via attention_args, by default 0.2
    glu_ff : bool, optional
        whether to use gated linear feedforward network, by default True
    activation : Literal["relu", "gelu", "swish"], optional
        activation, by default "swish"
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        **mod_kwargs,
    ):
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=False, **mod_kwargs)
        return x


class TransformerDecoder(Transformer):
    """Transformer Decoder network chaining multiple transformer layers
    Equal to Transformer, only difference is that Causal = True is automatically decided in forward pass

    Parameters
    ----------
    depth : int
        number of transformer blocks to use
    dim : int
        input and output hidden dimension of x
    nh : int
        number of heads, dim should be divisible by this number
    attentiontype : Literal["vanilla", "random", "window"], optional
        attention operator, by default "vanilla"
    attention_args : dict, optional
        args passed to the attention operator init, by default {}
    plugintype : Literal["none", "sinusoidal", "learned", "learnedcont", "rotary", "ALiBi", "DPB", "XL"], optional
        positional bias plugin, by default "none"
    plugin_args : dict, optional
        arguments passed to positional bias init, by default {}
    only_apply_plugin_at_first : bool, optional
        only apply positional bias at the first layer, by default False
    dropout : float, optional
        dropout in feedforward layers. Take note that attention matrix dropout is controlled via attention_args, by default 0.2
    glu_ff : bool, optional
        whether to use gated linear feedforward network, by default True
    activation : Literal["relu", "gelu", "swish"], optional
        activation, by default "swish"
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(
        self,
        x: torch.Tensor,
        pos: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        **mod_kwargs,
    ):
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=True, **mod_kwargs)
        return x


class Aggregator(nn.Module):
    """Aggregator module. Can be used to get a single vector from a sequence-valued input.

    Parameters
    ----------
    method : Literal["mean", "max", "cls"], optional
        aggregation method, by default "max"
    """

    def __init__(self, method: Literal["mean", "max", "cls"] = "max"):
        super().__init__()
        assert method in ["mean", "max", "cls"]
        self.method = method

    def forward(
        self, x: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        x : torch.Tensor
            (B, *, L, H)
        mask : Optional[torch.Tensor], optional
            (B, *, L), by default None

        Returns
        -------
        torch.Tensor
            (B, *, H)
        """
        if mask is not None:
            assert mask.dtype == torch.bool
            x = x * mask.unsqueeze(-1)
        if self.method == "mean":
            return x.sum(-2) / (
                mask.sum(-1, keepdim=True) if mask is not None else x.shape[-2]
            )
        elif self.method == "max":
            return x.max(-2).values
        elif self.method == "cls":
            return x[..., 0, :]
